/** \file app.c
 *  \brief Main-Functions.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <gnome.h>
#include <gtk/gtkgl.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <glade/glade.h>

#include "app.h"
#include "cam_setup.h"
#include "cam.h"
#include "cam_imgload.h"
#include "cnc_output.h"
#include "cam_object.h"

extern void on_glarea_realize (GtkWidget *widget, gpointer user_data);
extern gboolean on_glarea_expose_event (GtkWidget *widget, GdkEventExpose *event, gpointer user_data);

//! update the Info-Text displayed in the GUI(Statusbar)
void app__set_infobar (gchar *text_str) {
	g_mutex_lock(app->ViewMutex);
	strcpy(app->infotext, text_str);
	app->redraw = TRUE;
	g_mutex_unlock(app->ViewMutex);
}

void app__set_progress (gint percent) {
	g_mutex_lock(app->ViewMutex);
	app->progress = percent;
	app->redraw = TRUE;
	g_mutex_unlock(app->ViewMutex);
}

//! update the Size-Info displayed in the GUI(Viewbar)
void app__set_sizeinfo (void) {
	gchar *text_str;
	if ((text_str = malloc(20)) != NULL) {
		sprintf(text_str, "Size: %0.2f MB", (gfloat)(CAM__GetTotalMapSize() / 1024.0 / 1024.0));
		gtk_label_set_text(GTK_LABEL(glade_xml_get_widget(app->xml, "viewbar_hmsize")), text_str);
		free(text_str);
	}
}

//! shows a Error-Dialog
void app__show_error (gchar *title_str, gchar *text_str) {
	fprintf(stderr, "ERROR: %s\n", title_str);
	fprintf(stderr, "%s\n\n", text_str);
	app__set_infobar(text_str);
	gtk_window_set_title(GTK_WINDOW(glade_xml_get_widget(app->xml, "errordialog")), title_str);
	gtk_label_set_text(GTK_LABEL(glade_xml_get_widget(app->xml, "errordialog_message")), text_str);
	gtk_widget_show(glade_xml_get_widget(app->xml, "errordialog"));
}

//! shows an Info-Dialog
void app__show_info (gchar *title_str, gchar *text_str) {
	fprintf(stderr, "INFO: %s\n", title_str);
	fprintf(stderr, "%s\n\n", text_str);
	app__set_infobar(text_str);
	gtk_window_set_title(GTK_WINDOW(glade_xml_get_widget(app->xml, "infodialog")), title_str);
	gtk_label_set_text(GTK_LABEL(glade_xml_get_widget(app->xml, "infodialog_message")), text_str);
	gtk_widget_show(glade_xml_get_widget(app->xml, "infodialog"));
}

//! Draw the HeightMapView to the Screen
void app__view_redraw (void) {
	GdkColor color;
	gint x;
	gint y;
	gint drawable_width;
	gint drawable_height;
	gint z;
	if (app != NULL && app->HeightMapView != NULL) {
		app__set_sizeinfo();
		drawable_width = app->draw_widget->allocation.width;
		drawable_height = app->draw_widget->allocation.height;
		color.red = color.green = color.blue = 65535;
		gdk_gc_set_rgb_fg_color(app->gc, &color);
		gdk_draw_rectangle(app->draw_area, app->gc, TRUE, 0, 0, drawable_width, drawable_height);
		for (x = 0; x < drawable_width && (gint)((gfloat)x / app->hm_zoom) < app->HeightMapView->width; x++) {
			for (y = 0; y < drawable_height && (gint)((gfloat)y / app->hm_zoom) < app->HeightMapView->height; y++) {
				//color.red = color.green = color.blue = CAM__CheckPixel(app->HeightMapView, (gint)((gfloat)x / app->hm_zoom), (gint)((gfloat)y / app->hm_zoom));
				z = app->HeightMapView->data[((gint)((gfloat)y / app->hm_zoom) * app->HeightMapView->width + (gint)((gfloat)x / app->hm_zoom))];
				if (z == 65535) {
					color.red = 65535;
					color.green = 0;
					color.blue = 0;
				} else {
					z = (gint)((gfloat)z * (65535.0 / (gfloat)(app->CamSet->z_start)));
					z = 65535 - z;
					color.red = color.green = color.blue = z;
				}
				gdk_gc_set_rgb_fg_color(app->gc, &color);
				gdk_draw_point(app->draw_area, app->gc, x, (drawable_height - y));
			}
		}
		gtk_ruler_set_range(GTK_RULER(glade_xml_get_widget(app->xml, "draw_xruler")),
			-3.0, (gdouble)((gfloat)(drawable_width + 20) / app->hm_zoom) / 10.0 * app->CamSet->zoom,
			10.0, (gdouble)((gfloat)(drawable_width + 20) / app->hm_zoom) / 10.0 * app->CamSet->zoom
		);
		gtk_ruler_set_range(GTK_RULER(glade_xml_get_widget(app->xml, "draw_yruler")),
			(gdouble)((gfloat)(drawable_height + 20) / app->hm_zoom) / 10.0 * app->CamSet->zoom, -3.0,
			10.0, (gdouble)((gfloat)(drawable_height + 20) / app->hm_zoom) / 10.0 * app->CamSet->zoom
		);
//		on_glarea_realize(app->gl_area, NULL);
	}
}

//! copy Heightmap to HeightMapView and call app__view_redraw()
void app__showmap(CAM__HeightMap *HeightMap) {
	if (HeightMap != NULL && HeightMap != app->HeightMapView) {
		CAM__Heightmap_delete(app->HeightMapView);
		app->HeightMapView = CAM__HeightMap_Copy(HeightMap);
		app__view_redraw();
	}
}

//! load an Image as a new Heightmap
void app__load_file (gchar *filename, gint z_stop, gint z_start, gboolean use_rg, gboolean invert) {
		app__set_infobar(_("loading Heightmap..."));
		CAM__Heightmap_delete(app->HeightMapSource);
		app->HeightMapSource = CAM__Heightmap_LoadImage(filename, z_stop, z_start, use_rg, invert);
		app->CamSet->z_start = z_start;
		app->CamSet->z_stop = z_stop;
		app__showmap(app->HeightMapSource);
		sprintf(app->infotext, "%s = %s", _("new Heightmap"), gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(app->filedialog)));
		app__set_infobar(app->infotext);
		sprintf(app->infotext, "%s (%s)", _("Heigmap->CNC"), gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(app->filedialog)));
		gtk_window_set_title(GTK_WINDOW(app->window), app->infotext);
		gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(app->filedialog), filename);
}

gboolean app__update (gpointer data) {
	g_mutex_lock(app->ViewMutex);
	if (app->gl_area != NULL) {
		gdk_window_invalidate_rect(app->gl_area->window, &app->gl_area->allocation, FALSE);
		gdk_window_process_updates(app->gl_area->window, TRUE);
		on_glarea_expose_event(NULL, NULL, NULL);
	}
	if (app->redraw == TRUE) {
		gtk_statusbar_pop(app->statusbar, 1);
		gtk_statusbar_push(app->statusbar, 1, app->infotext);
		gchar temp_str[10];
		sprintf(temp_str, "%i %%", app->progress);
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(glade_xml_get_widget(app->xml, "cnc2d_progressbar")), (gdouble)app->progress / 100.0);
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR(glade_xml_get_widget(app->xml, "cnc2d_progressbar")), temp_str);
		app->redraw = FALSE;
	}
	g_mutex_unlock(app->ViewMutex);
	return TRUE;
}

//! Callback -- Area / Drop
void on_cnc2d_drag_data_received (GtkWidget *widget, GdkDragContext *dc, gint x, gint y, GtkSelectionData *selection_data, guint info, guint t, gpointer data) {
	const gchar* file = strtok((gchar*)selection_data->data,"\n");
	gchar filename[2048];
	while(file != NULL) {
		if (strncmp(file, "file://", 7) == 0
			&& (strstr(file, ".png\0") != 0
			|| strstr(file, ".bmp\0") != 0
			|| strstr(file, ".tiff\0") != 0
			|| strstr(file, ".tif\0") != 0
			|| strstr(file, ".jpeg\0") != 0
			|| strstr(file, ".jpg\0") != 0
			|| strstr(file, ".xpm\0") != 0)
		) {
			strcpy(filename, file + 7);
			filename[strlen(filename) - 1] = 0;
			fprintf(stderr, "Import Image: %s\n", filename);
			app__load_file(filename, 0, 2000, FALSE, FALSE);
			break;
		} else if (strncmp(file, "file://", 7) == 0
			&& (strstr(file, ".hpgl\0") != 0
			|| strstr(file, ".plt\0") != 0)
		) {
			strcpy(filename, file + 7);
			filename[strlen(filename) - 1] = 0;
			fprintf(stderr, "Import HPGL: %s\n", filename);
			break;
		} else {
			fprintf(stderr, "Unknown: %s\n", file);
			break;
		}
		file = strtok(NULL, "\n");
	}
}

//! create a new application-context
Cnc2D *app__new (void) {
	Cnc2D *app;
	app = malloc(sizeof(Cnc2D));
	app->xml = glade_xml_new(PACKAGE_DATA_DIR"/"PACKAGE"/glade/"PACKAGE".glade", NULL, NULL);
	glade_xml_signal_autoconnect(app->xml);
	app->statusbar = GTK_STATUSBAR(glade_xml_get_widget(app->xml, "statusbar"));
	app->window = glade_xml_get_widget(app->xml, "cnc2d");
	app->draw_area = glade_xml_get_widget(app->xml, "draw_area")->window;
	app->draw_widget = glade_xml_get_widget(app->xml, "draw_area");
	app->gc = gdk_gc_new(app->draw_area);
	app->infotext = malloc(1024);
	app->progress = 50;
	app->filename = NULL;
	app->hm_zoom = (gfloat)(gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(app->xml, "hm_options_zoom_spin"))));
	app->gl_mode = 0;
	app->gl_zoom = (gfloat)(gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(app->xml, "gl_options_zoom_spin"))));
	app->gl_zzoom = (gfloat)(gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(app->xml, "gl_options_zzoom_spin"))));

	// Filedialog (Save)
	app->filesavedialog = glade_xml_get_widget(app->xml, "filesavedialog");
	app->img_filter2 = gtk_file_filter_new();
	gtk_file_filter_add_pixbuf_formats(app->img_filter2);
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(app->filesavedialog), app->img_filter2);
	gtk_window_set_title(GTK_WINDOW(app->filesavedialog), _("save Heigmap"));

	// Filedialog (Load)
	app->filedialog = glade_xml_get_widget(app->xml, "filedialog");
	app->img_filter = gtk_file_filter_new();
	gtk_file_filter_add_pixbuf_formats(app->img_filter);
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(app->filedialog), app->img_filter);
	gtk_window_set_title(GTK_WINDOW(app->filedialog), _("load Heigmap"));
	gtk_window_set_title(GTK_WINDOW(app->window), _("Heigmap->CNC"));
	gtk_widget_show(app->window);

	CNC__Init();
	app->CamSet = camsetup__init();
	app->HeightMapSource = NULL;
	app->HeightMapView = NULL;

	app->ViewMutex = g_mutex_new();
	app->redraw = TRUE;
	g_timeout_add(100, app__update, NULL);

	enum {
		TARGET_URI_LIST,
		TARGET_STRING,
		TARGET_COUNT
	};
	static GtkTargetEntry targetentries[] = {
		{"text/uri-list", 0, TARGET_URI_LIST},
		{"text/plain", 0, TARGET_STRING}
	};
	gtk_drag_dest_set(GTK_WIDGET(app->window), GTK_DEST_DEFAULT_ALL, targetentries, TARGET_COUNT, GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK);

	return app;
}
