/** \file cam.c
 *  \brief CAM-Functions to generate CNC-Output from Heightmaps
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <gnome.h>
#include <glade/glade.h>
#include <GL/gl.h>

#include "app.h"
#include "cam_setup.h"
#include "cam.h"
#include "cnc_output.h"

gint CAM__Heightmap_num_maps = 0; //! number of allocated Heightmaps
gdouble CAM__Heightmap_mapsize_total = 0;  //! total size of all allocated Heightmaps

//! returns the diff of 2 values
gint CAM__GetDiff(gint val1, gint val2) {
	if (val1 > val2) {
		return(val1 - val2);
	} else if (val1 < val2) {
		return(val2 - val1);
	}
	return 0;
}

//! returns the smaller value of the 2 vars
gint CAM__SetMin(gint val1, gint val2) {
	if (val1 < val2) {
		return val1;
	} else {
		return val2;
	}
}

//! returns the larger value of the 2 vars
gint CAM__SetMax(gint val1, gint val2) {
	if (val1 > val2) {
		return val1;
	} else {
		return val2;
	}
}

//! faster macro of CAM__SetMax(val1, val2) / sets val2 to the larger value
#define _CAM__SetMax(val1, val2) {\
	if (val1 > val2) {\
		val2 = val1;\
	}\
}\

//! returns the total size of all allocated Heightmaps
gdouble CAM__GetTotalMapSize(void) {
	return CAM__Heightmap_mapsize_total;
}

//! returns the number of allocated Heightmaps
gint CAM__GetTotalMapNum(void) {
	return CAM__Heightmap_num_maps;
}

//! get the Height of a Position(x,y) in a HeightMap
gint CAM__HeightMap_CheckPixel(CAM__HeightMap *HeightMap, gint x, gint y) {
    return HeightMap->data[(y * HeightMap->width) + x];
}

//! returns the largest Heightvalue Mask by an circle(x,y,r)
gint CAM__HeightMap_CheckCircleMax(CAM__HeightMap *HeightMap, gint x0, gint y0, gint radius_max) {
	gint f;
	gint ddF_x;
	gint ddF_y;
	gint x;
	gint y;
	gint radius;
	gint max = 0;
	for (radius = 0; radius <= radius_max; radius++) {
		f = 1 - radius;
		ddF_x = 0;
		ddF_y = -2 * radius;
		x = 0;
		y = radius;
		_CAM__SetMax(HeightMap->data[((y0 + radius) * HeightMap->width) + x0], max);
		_CAM__SetMax(HeightMap->data[((y0 - radius) * HeightMap->width) + x0], max);
		_CAM__SetMax(HeightMap->data[(y0 * HeightMap->width) + (x0 + radius)], max);
		_CAM__SetMax(HeightMap->data[(y0 * HeightMap->width) + (x0 - radius)], max);
		while(x < y) {
			if(f >= 0) {
				y--;
				ddF_y += 2;
				f += ddF_y;
			}
			x++;
			ddF_x += 2;
			f += ddF_x + 1;
			_CAM__SetMax(HeightMap->data[((y0 + y) * HeightMap->width) + (x0 + x)], max);
			_CAM__SetMax(HeightMap->data[((y0 + y) * HeightMap->width) + (x0 - x)], max);
			_CAM__SetMax(HeightMap->data[((y0 - y) * HeightMap->width) + (x0 + x)], max);
			_CAM__SetMax(HeightMap->data[((y0 - y) * HeightMap->width) + (x0 - x)], max);
			_CAM__SetMax(HeightMap->data[((y0 + x) * HeightMap->width) + (x0 + y)], max);
			_CAM__SetMax(HeightMap->data[((y0 + x) * HeightMap->width) + (x0 - y)], max);
			_CAM__SetMax(HeightMap->data[((y0 - x) * HeightMap->width) + (x0 + y)], max);
			_CAM__SetMax(HeightMap->data[((y0 - x) * HeightMap->width) + (x0 - y)], max);
		}
	}
	return max;
}

//! returns the smallest Heightvalue mask by an circle(x,y,r)
gint CAM__HeightMap_CheckCircleMin(CAM__HeightMap *HeightMap, gint x0, gint y0, gint radius_max) {
	gint f;
	gint ddF_x;
	gint ddF_y;
	gint x;
	gint y;
	gint radius;
	gint min = 65535;
	for (radius = 0; radius <= radius_max; radius++) {
		f = 1 - radius;
		ddF_x = 0;
		ddF_y = -2 * radius;
		x = 0;
		y = radius;
		min = CAM__SetMin(HeightMap->data[((y0 + radius) * HeightMap->width) + x0], min);
		min = CAM__SetMin(HeightMap->data[((y0 - radius) * HeightMap->width) + x0], min);
		min = CAM__SetMin(HeightMap->data[(y0 * HeightMap->width) + (x0 + radius)], min);
		min = CAM__SetMin(HeightMap->data[(y0 * HeightMap->width) + (x0 - radius)], min);
		while(x < y) {
			if(f >= 0) {
				y--;
				ddF_y += 2;
				f += ddF_y;
			}
			x++;
			ddF_x += 2;
			f += ddF_x + 1;
			min = CAM__SetMin(HeightMap->data[((y0 + y) * HeightMap->width) + (x0 + x)], min);
			min = CAM__SetMin(HeightMap->data[((y0 + y) * HeightMap->width) + (x0 - x)], min);
			min = CAM__SetMin(HeightMap->data[((y0 - y) * HeightMap->width) + (x0 + x)], min);
			min = CAM__SetMin(HeightMap->data[((y0 - y) * HeightMap->width) + (x0 - x)], min);
			min = CAM__SetMin(HeightMap->data[((y0 + x) * HeightMap->width) + (x0 + y)], min);
			min = CAM__SetMin(HeightMap->data[((y0 + x) * HeightMap->width) + (x0 - y)], min);
			min = CAM__SetMin(HeightMap->data[((y0 - x) * HeightMap->width) + (x0 + y)], min);
			min = CAM__SetMin(HeightMap->data[((y0 - x) * HeightMap->width) + (x0 - y)], min);
		}
	}
	return min;
}

//! returns 0 if all Heightvalues mask by an circle(x,y,r) != 65535
gint CAM__HeightMap_CheckCircle(CAM__HeightMap *HeightMap, gint x0, gint y0, gint radius2, gint radius1) {
	gint f;
	gint ddF_x;
	gint ddF_y;
	gint x;
	gint y;
	gint radius;
	for (radius = radius2; radius >= radius1; radius--) {
		f = 1 - radius;
		ddF_x = 0;
		ddF_y = -2 * radius;
		x = 0;
		y = radius;
		if (HeightMap->data[((y0 + radius) * HeightMap->width) + x0] == 65535) {
			return 1;
		}
		if (HeightMap->data[((y0 - radius) * HeightMap->width) + x0] == 65535) {
			return 1;
		}
		if (HeightMap->data[(y0 * HeightMap->width) + (x0 + radius)] == 65535) {
			return 1;
		}
		if (HeightMap->data[(y0 * HeightMap->width) + (x0 - radius)] == 65535) {
			return 1;
		}
		while(x < y) {
			if(f >= 0) {
				y--;
				ddF_y += 2;
				f += ddF_y;
			}
			x++;
			ddF_x += 2;
			f += ddF_x + 1;
			if (HeightMap->data[((y0 + y) * HeightMap->width) + (x0 + x)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 + y) * HeightMap->width) + (x0 - x)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 - y) * HeightMap->width) + (x0 + x)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 - y) * HeightMap->width) + (x0 - x)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 + x) * HeightMap->width) + (x0 + y)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 + x) * HeightMap->width) + (x0 - y)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 - x) * HeightMap->width) + (x0 + y)] == 65535) {
				return 1;
			}
			if (HeightMap->data[((y0 - x) * HeightMap->width) + (x0 - y)] == 65535) {
				return 1;
			}
		}
	}
	return 0;
}

//! CNC-Liner
gint CAM__HeightMap_CNC_Liner(CAM__HeightMap *HeightMap, gint min_x, gint min_y, gint max_x, gint max_y, CamSetup *CamSet) {
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom);
	gint test_x = 0;
	gint test_y = 0;
	gint test2_x = 0;
	gint test2_y = 0;
	gint last_x = 0;
	gint last_y = 0;
	gint first_x = 0;
	gint first_y = 0;
	gint dir_x = 0;
	gint dir_y = 0;
	gint ldir_x = 0;
	gint ldir_y = 0;
	gint y_addr = 0;
	gint diff = 1;
	gint depth = 65535;
	gint last_depth = 0;
	gint find_val = 65535;
	if (min_x < 0) {min_x = 0;}
	if (min_y < 0) {min_y = 0;}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	for (test2_y = min_y; CNC__OutActive == 1 && test2_y < max_y; test2_y++) {
		test_y = test2_y;
		y_addr = (test_y * HeightMap->width);
		for (test2_x = min_x; CNC__OutActive == 1 && test2_x < max_x; test2_x++) {
			test_x = test2_x;
			if (HeightMap->data[(test_y * HeightMap->width) + test_x] != find_val) {
				depth = HeightMap->data[(test_y * HeightMap->width) + test_x];
				HeightMap->data[(test_y * HeightMap->width) + test_x] = 65535;
				first_x = test_x;
				first_y = test_y;
				last_x = test_x;
				last_y = test_y;
				last_depth = depth;
				ldir_x = 0;
				ldir_y = 0;
				CNC__Go_XY(test_x, test_y);
				CNC__DrillSetSpeed(CamSet->axis_z[CamSet->tool].tool.speed);
//				CNC__Go_Z(depth);
				CNC__Go_Z(0);
				while (CNC__OutActive == 1) {
					if (test_x + diff < max_x && HeightMap->data[(test_y * HeightMap->width) + (test_x + diff)] != find_val) {
						test_x += diff;
						dir_x = 1;
						dir_y = 0;
					} else if (test_x - diff >= 0 && HeightMap->data[(test_y * HeightMap->width) + (test_x - diff)] != find_val) {
						test_x -= diff;
						dir_x = -1;
						dir_y = 0;
					} else if (test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + test_x] != find_val) {
						test_y += diff;
						dir_x = 0;
						dir_y = 1;
					} else if (test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + test_x] != find_val) {
						test_y -= diff;
						dir_x = 0;
						dir_y = -1;
					} else if (test_x + diff < max_x && test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + (test_x + diff)] != find_val) {
						test_x += diff;
						test_y += diff;
						dir_x = 1;
						dir_y = 1;
					} else if (test_x + diff < max_x && test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + (test_x + diff)] != find_val) {
						test_x += diff;
						test_y -= diff;
						dir_x = 1;
						dir_y = -1;
					} else if (test_x - diff >= 0 && test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + (test_x - diff)] != find_val) {
						test_x -= diff;
						test_y -= diff;
						dir_x = -1;
						dir_y = -1;
					} else if (test_x - diff >= 0 && test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + (test_x - diff)] != find_val) {
						test_x -= diff;
						test_y += diff;
						dir_x = -1;
						dir_y = 1;
					} else {
						break;
					}
					depth = HeightMap->data[(test_y * HeightMap->width) + test_x];
					//fprintf(stderr, "### %i %i,%i\n", depth, test_x, test_y);
					HeightMap->data[(test_y * HeightMap->width) + test_x] = 65535;
//					if (ldir_x != dir_x || ldir_y != dir_y || last_depth != depth) {
					if (ldir_x != dir_x || ldir_y != dir_y) {
						CNC__Go_XY(last_x, last_y);
						if (depth < last_depth) {
							CNC__Go_XY(test_x, test_y);
							CNC__Go_Z(0);
						} else {
							CNC__Go_Z(0);
							CNC__Go_XY(test_x, test_y);
						}
						//fprintf(stderr, "Depth: %i\n", depth);
					}
					last_depth = depth;
					last_x = test_x;
					last_y = test_y;
					ldir_x = dir_x;
					ldir_y = dir_y;
				}
				CNC__Go_XY(last_x, last_y);
				if (CAM__GetDiff(first_x, test_x) < radius && CAM__GetDiff(first_y, test_y) < radius) {
					CNC__Go_XY(first_x, first_y);
				}
				CNC__Go_Z(CamSet->z_start + CamSet->z_over);
				CNC__DrillSetSpeed(0);
			}
		}
	}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	return 0;
}

//! CNC-Liner
gint CAM__HeightMap_CNC_Liner3D(CAM__HeightMap *HeightMap, gint min_x, gint min_y, gint max_x, gint max_y, CamSetup *CamSet) {
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom);
	gint test_x = 0;
	gint test_y = 0;
	gint test2_x = 0;
	gint test2_y = 0;
	gint last_x = 0;
	gint last_y = 0;
	gint first_x = 0;
	gint first_y = 0;
	gint dir_x = 0;
	gint dir_y = 0;
	gint ldir_x = 0;
	gint ldir_y = 0;
	gint y_addr = 0;
	gint diff = 1;
	gint depth = 65535;
	gint last_depth = 0;
	gint find_val = 65535;
	if (min_x < 0) {min_x = 0;}
	if (min_y < 0) {min_y = 0;}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	for (test2_y = min_y; CNC__OutActive == 1 && test2_y < max_y; test2_y++) {
		test_y = test2_y;
		y_addr = (test_y * HeightMap->width);
		for (test2_x = min_x; CNC__OutActive == 1 && test2_x < max_x; test2_x++) {
			test_x = test2_x;
			if (HeightMap->data[(test_y * HeightMap->width) + test_x] != find_val) {
				depth = HeightMap->data[(test_y * HeightMap->width) + test_x];
				HeightMap->data[(test_y * HeightMap->width) + test_x] = 65535;
				first_x = test_x;
				first_y = test_y;
				last_x = test_x;
				last_y = test_y;
				last_depth = depth;
				ldir_x = 0;
				ldir_y = 0;
				CNC__Go_XY(test_x, test_y);
				CNC__DrillSetSpeed(CamSet->axis_z[CamSet->tool].tool.speed);
				CNC__Go_Z(depth);
				while (CNC__OutActive == 1) {
					if (test_x + diff < max_x && HeightMap->data[(test_y * HeightMap->width) + (test_x + diff)] != find_val) {
						test_x += diff;
						dir_x = 1;
						dir_y = 0;
					} else if (test_x - diff >= 0 && HeightMap->data[(test_y * HeightMap->width) + (test_x - diff)] != find_val) {
						test_x -= diff;
						dir_x = -1;
						dir_y = 0;
					} else if (test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + test_x] != find_val) {
						test_y += diff;
						dir_x = 0;
						dir_y = 1;
					} else if (test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + test_x] != find_val) {
						test_y -= diff;
						dir_x = 0;
						dir_y = -1;
					} else if (test_x + diff < max_x && test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + (test_x + diff)] != find_val) {
						test_x += diff;
						test_y += diff;
						dir_x = 1;
						dir_y = 1;
					} else if (test_x + diff < max_x && test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + (test_x + diff)] != find_val) {
						test_x += diff;
						test_y -= diff;
						dir_x = 1;
						dir_y = -1;
					} else if (test_x - diff >= 0 && test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + (test_x - diff)] != find_val) {
						test_x -= diff;
						test_y -= diff;
						dir_x = -1;
						dir_y = -1;
					} else if (test_x - diff >= 0 && test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + (test_x - diff)] != find_val) {
						test_x -= diff;
						test_y += diff;
						dir_x = -1;
						dir_y = 1;
					} else {
						break;
					}
					depth = HeightMap->data[(test_y * HeightMap->width) + test_x];
					//fprintf(stderr, "### %i %i,%i\n", depth, test_x, test_y);
					HeightMap->data[(test_y * HeightMap->width) + test_x] = 65535;
//					if (ldir_x != dir_x || ldir_y != dir_y || last_depth != depth) {
					if (ldir_x != dir_x || ldir_y != dir_y) {
						CNC__Go_XY(last_x, last_y);
						if (depth < last_depth) {
							CNC__Go_XY(test_x, test_y);
							CNC__Go_Z(depth);
						} else {
							CNC__Go_Z(depth);
							CNC__Go_XY(test_x, test_y);
						}
						//fprintf(stderr, "Depth: %i\n", depth);
					}
					last_depth = depth;
					last_x = test_x;
					last_y = test_y;
					ldir_x = dir_x;
					ldir_y = dir_y;
				}
				CNC__Go_XY(last_x, last_y);
				if (CAM__GetDiff(first_x, test_x) < radius && CAM__GetDiff(first_y, test_y) < radius) {
					CNC__Go_XY(first_x, first_y);
				}
				CNC__Go_Z(CamSet->z_start + CamSet->z_over);
				CNC__DrillSetSpeed(0);
			}
		}
	}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	return 0;
}

//! CNC-Filler (Old-Version)
gint CAM__HeightMap_CNC_Liner_old(CAM__HeightMap *HeightMap, gint min_x, gint min_y, gint max_x, gint max_y, CamSetup *CamSet) {
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom);
	gint test_x = 0;
	gint test_y = 0;
	gint test2_x = 0;
	gint test2_y = 0;
	gint last_x = 0;
	gint last_y = 0;
	gint first_x = 0;
	gint first_y = 0;
	gint dir_x = 0;
	gint dir_y = 0;
	gint ldir_x = 0;
	gint ldir_y = 0;
	gint y_addr = 0;
	gint diff = 1;
	gint find_val = 65535;
	if (min_x < 0) {min_x = 0;}
	if (min_y < 0) {min_y = 0;}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	for (test2_y = min_y; CNC__OutActive == 1 && test2_y < max_y; test2_y++) {
		test_y = test2_y;
		y_addr = (test_y * HeightMap->width);
		for (test2_x = min_x; CNC__OutActive == 1 && test2_x < max_x; test2_x++) {
			test_x = test2_x;
			if (HeightMap->data[(test_y * HeightMap->width) + test_x] == find_val) {
				HeightMap->data[(test_y * HeightMap->width) + test_x] = 65535 / 2;
				CNC__Go_XY(test_x, test_y);
				first_x = test_x;
				first_y = test_y;
				last_x = test_x;
				last_y = test_y;
				ldir_x = 0;
				ldir_y = 0;
				CNC__DrillSetSpeed(CamSet->axis_z[CamSet->tool].tool.speed);
				CNC__Go_Z(CamSet->z_step);
				while (CNC__OutActive == 1) {
					if (test_x + diff < max_x && HeightMap->data[(test_y * HeightMap->width) + (test_x + diff)] == find_val) {
						test_x += diff;
						dir_x = 1;
						dir_y = 0;
					} else if (test_x - diff >= 0 && HeightMap->data[(test_y * HeightMap->width) + (test_x - diff)] == find_val) {
						test_x -= diff;
						dir_x = -1;
						dir_y = 0;
					} else if (test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + test_x] == find_val) {
						test_y += diff;
						dir_x = 0;
						dir_y = 1;
					} else if (test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + test_x] == find_val) {
						test_y -= diff;
						dir_x = 0;
						dir_y = -1;
					} else if (test_x + diff < max_x && test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + (test_x + diff)] == find_val) {
						test_x += diff;
						test_y += diff;
						dir_x = 1;
						dir_y = 1;
					} else if (test_x + diff < max_x && test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + (test_x + diff)] == find_val) {
						test_x += diff;
						test_y -= diff;
						dir_x = 1;
						dir_y = -1;
					} else if (test_x - diff >= 0 && test_y - diff >= 0 && HeightMap->data[((test_y - diff) * HeightMap->width) + (test_x - diff)] == find_val) {
						test_x -= diff;
						test_y -= diff;
						dir_x = -1;
						dir_y = -1;
					} else if (test_x - diff >= 0 && test_y + diff < max_y && HeightMap->data[((test_y + diff) * HeightMap->width) + (test_x - diff)] == find_val) {
						test_x -= diff;
						test_y += diff;
						dir_x = -1;
						dir_y = 1;
					} else {
						break;
					}
					HeightMap->data[(test_y * HeightMap->width) + test_x] = 65535 / 2;
					if (ldir_x != dir_x || ldir_y != dir_y) {
						CNC__Go_XY(last_x, last_y);
						CNC__Go_XY(test_x, test_y);
					}
					last_x = test_x;
					last_y = test_y;
					ldir_x = dir_x;
					ldir_y = dir_y;
				}
				CNC__Go_XY(last_x, last_y);
				if (CAM__GetDiff(first_x, test_x) < radius && CAM__GetDiff(first_y, test_y) < radius) {
					CNC__Go_XY(first_x, first_y);
				}
				CNC__Go_Z(CamSet->z_start + CamSet->z_over);
				CNC__DrillSetSpeed(0);
			}
		}
	}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	return 0;
}

//! CNC-Soft-3D
gint CAM__HeightMap_CNC_Soft3D(CAM__HeightMap *HeightMap, CamSetup *CamSet) {
	gint x;
	gint y;
	gint value;
	gint last;
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom);
	gint start_z = CamSet->z_start + CamSet->z_over;
	last = start_z;
	CNC__DrillSetSpeed(CamSet->axis_z[CamSet->tool].tool.speed);
	//fprintf(stderr, "Soft3D: start (%i, %i, %i)\n", CamSet->z_start, CamSet->z_stop, CamSet->z_steps);
	CNC__Go_Z(start_z);
	for (y = radius; CNC__OutActive == 1 && y < HeightMap->height - radius; y += radius / 2) {
		for (x = radius; CNC__OutActive == 1 && x < HeightMap->width - radius; x += radius / 2) {
			value = CAM__SetMax(CamSet->z_step, CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius));
			if (value > CamSet->z_start) {
				value = CamSet->z_start;
			}
			if (value < last) {
				CNC__Go_XY(x, y);
				CNC__Go_Z(value);
			} else {
				CNC__Go_Z(value);
				CNC__Go_XY(x, y);
			}
			last = value;
		}
		y += radius  / 2;
		if (y < HeightMap->height - radius) {
			for (x = HeightMap->width - 1 - radius; CNC__OutActive == 1 && x >= radius; x -= radius / 2) {
				value = CAM__SetMax(CamSet->z_step, CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius));
				if (value > CamSet->z_start) {
					value = CamSet->z_start;
				}
				if (value < last) {
					CNC__Go_XY(x, y);
					CNC__Go_Z(value);
				} else {
					CNC__Go_Z(value);
					CNC__Go_XY(x, y);
				}
				last = value;
			}
		}
	}
	CNC__Go_Z(CamSet->z_start + CamSet->z_over);
	CNC__DrillSetSpeed(0);
	return 0;
}

//! Inverts the Values of a Heightmap
void CAM__HeightMap_Invert(CAM__HeightMap *HeightMap) {
	gint n;
	for (n = 0; n < (HeightMap->height * HeightMap->width); n++) {
		HeightMap->data[n] = 65535 - HeightMap->data[n];
	}
}

//! delete an allocated Heightmap
void CAM__Heightmap_delete(CAM__HeightMap *HeightMap) {
	if (HeightMap != NULL) {
		CAM__Heightmap_mapsize_total -= (gdouble)(sizeof(CAM__HeightMap) + (HeightMap->width * HeightMap->height * sizeof(gint)));
		if (HeightMap->data != NULL) {
			free(HeightMap->data);
			HeightMap->data = NULL;
		}
		free(HeightMap);
		HeightMap = NULL;
		CAM__Heightmap_num_maps--;
	}
}

//! returns a new allocated Heightmap
CAM__HeightMap *CAM__HeightMap_create (gint width, gint height) {
	CAM__HeightMap *HeightMap_New;
	HeightMap_New = malloc(sizeof(CAM__HeightMap));
	if (HeightMap_New == NULL) {
			fprintf(stderr, "Can not alloc Memory for a new Heightmap\n");
			return NULL;
	}
	HeightMap_New->width = width;
	HeightMap_New->height = height;
	HeightMap_New->use_rg = FALSE;
	HeightMap_New->data = malloc(HeightMap_New->width * HeightMap_New->height * sizeof(gint));
	if (HeightMap_New->data == NULL) {
			fprintf(stderr, "Can not alloc Memory for the Heightmap->data (%0.2fMB)\n", (gfloat)HeightMap_New->width * (gfloat)HeightMap_New->height * (gfloat)sizeof(gint) / 1024.0 / 1024.0);
			free(HeightMap_New);
			return NULL;
	}
	CAM__Heightmap_mapsize_total += (gdouble)(sizeof(CAM__HeightMap) + (HeightMap_New->width * HeightMap_New->height * sizeof(gint)));
	CAM__Heightmap_num_maps++;
	return HeightMap_New;
}

//! copy given Heightmap to a new Heightmap
CAM__HeightMap *CAM__HeightMap_Copy(CAM__HeightMap *HeightMap) {
	CAM__HeightMap *HeightMap_New;
//	gint n;
	if ((HeightMap_New = CAM__HeightMap_create(HeightMap->width, HeightMap->height)) != NULL) {
		memcpy(HeightMap_New->data, HeightMap->data, HeightMap->height * HeightMap->width * sizeof(gint));
//		for (n = 0; n < (HeightMap->height * HeightMap->width); n++) {
//			HeightMap_New->data[n] = HeightMap->data[n];
//		}
		return HeightMap_New;
	}
	return NULL;
}

//! copy given Heightmap to a new Heightmap
CAM__HeightMap *CAM__HeightMap_AddFrame(CAM__HeightMap *HeightMap, gint width) {
	CAM__HeightMap *HeightMap_New;
	gint x;
	gint y;
	gint n;
	if ((HeightMap_New = CAM__HeightMap_create(HeightMap->width + width + width, HeightMap->height + width + width)) == NULL) {
		return NULL;
	}
	for (n = 0; n < (HeightMap_New->height * HeightMap_New->width); n++) {
		HeightMap_New->data[n] = 65535;
	}
	for (y = 0; y < HeightMap->height; y++) {
		for (x = 0; x < HeightMap->width; x++) {
			HeightMap_New->data[((y + width) * HeightMap_New->width) + (x + width)] = HeightMap->data[(y * HeightMap->width) + x];
		}
	}
	return HeightMap_New;
}

//! copy given Heightmap to a new Heightmap (only Pixels are smaller then the given depth)
CAM__HeightMap *CAM__HeightMap_DepthCopy(CAM__HeightMap *HeightMap, gint depth) {
	CAM__HeightMap *HeightMap_New;
	gint n;
	if ((HeightMap_New = CAM__HeightMap_create(HeightMap->width, HeightMap->height)) != NULL) {
		for (n = 0; n < (HeightMap->height * HeightMap->width); n++) {
			if (HeightMap->data[n] <= depth) {
				HeightMap_New->data[n] = HeightMap->data[n];
			} else {
				HeightMap_New->data[n] = 65535;
			}
		}
		return HeightMap_New;
	}
	return NULL;
}

//! returns a new HeightMap of the Contour of the given one
CAM__HeightMap *CAM__HeightMap_Contour(CAM__HeightMap *HeightMap, CamSetup *CamSet) {
	CAM__HeightMap *HeightMap_New;
	gint x;
	gint y;
	gint n;
	gint value;
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom);
	if ((HeightMap_New = CAM__HeightMap_create(HeightMap->width, HeightMap->height)) == NULL) {
		fprintf(stderr, "CAM__HeightMap_Contour...failed\n");
		return NULL;
	}
	for (n = 0; n < (HeightMap_New->height * HeightMap_New->width); n++) {
		HeightMap_New->data[n] = 65535;
	}
	for (y = radius + 1; CNC__OutActive == 1 && y < HeightMap->height - radius - 1; y++) {
		for (x = radius + 1; CNC__OutActive == 1 && x < HeightMap->width - radius - 1; x++) {
			if (HeightMap->data[(y * HeightMap->width) + x] != 65535 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius + 1, radius + 1) != 0 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius, 1) == 0) {
				value = CAM__SetMax(CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius), CamSet->z_step);
				if (value > CamSet->z_start) {
					value = CamSet->z_start;
				}
				HeightMap_New->data[(y * HeightMap_New->width) + x] = value;
			} else {
				HeightMap_New->data[(y * HeightMap_New->width) + x] = 65535;
			}
		}
	}
	return HeightMap_New;
}

//! returns a new HeightMap of the Outlines of the given one
CAM__HeightMap *CAM__HeightMap_Outline(CAM__HeightMap *HeightMap, CamSetup *CamSet) {
	CAM__HeightMap *HeightMap_New;
	gint x;
	gint y;
	gint n;
	gint value;
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom);
	if ((HeightMap_New = CAM__HeightMap_create(HeightMap->width, HeightMap->height)) == NULL) {
		fprintf(stderr, "CAM__HeightMap_Outline...failed\n");
		return NULL;
	}
	for (n = 0; n < (HeightMap_New->height * HeightMap_New->width); n++) {
		HeightMap_New->data[n] = 65535;
	}
	for (y = radius + 1; CNC__OutActive == 1 && y < HeightMap->height - radius - 1; y++) {
		for (x = radius + 1; CNC__OutActive == 1 && x < HeightMap->width - radius - 1; x++) {
			if (HeightMap->data[(y * HeightMap->width) + x] != 65535 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius + 1, radius + 1) != 0 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius, 1) == 0) {
				value = CAM__SetMax(CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius), CamSet->z_step);
				if (value > CamSet->z_start) {
					value = CamSet->z_start;
				}
				HeightMap_New->data[(y * HeightMap_New->width) + x] = value;
			} else {
				HeightMap_New->data[(y * HeightMap_New->width) + x] = 65535;
			}
		}
	}
	return HeightMap_New;
}

//! CNC-Filler
gint CAM__HeightMap_CNC_Fill(CAM__HeightMap *HeightMap, CamSetup *CamSet) {
	gint x;
	gint y;
	gint value;
	gint last_value;
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom) + (gint)(((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / 2.0) / (gfloat)CamSet->zoom) + 1;
	gint start_z = CamSet->z_start + CamSet->z_over;
	last_value = start_z;
	CNC__Go_Z(start_z);
	CNC__Go_XY(radius + 1, radius + 1);
	CNC__DrillSetSpeed(CamSet->axis_z[CamSet->tool].tool.speed);
	for (y = radius + 1; CNC__OutActive == 1 && y < HeightMap->height - radius - 1; y += (radius / 2) + 1) {
		for (x = radius + 1; CNC__OutActive == 1 && x < HeightMap->width - radius - 1; x++) {
			if (HeightMap->data[(y * HeightMap->width) + x] != 65535 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius, radius) == 0 && (value = CAM__SetMax(CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius), CamSet->z_step)) != 65535) {
			} else {
				value = start_z;
			}
			if (value < last_value) {
				CNC__Go_XY(x, y);
				CNC__Go_Z(value);
			} else if (value > last_value) {
				CNC__Go_XY(x, y);
				CNC__Go_Z(start_z);
			}
			last_value = value;
		}
		if (last_value != start_z) {
			CNC__Go_XY(HeightMap->width - radius - 1 - 1, y);
			CNC__Go_Z(last_value);
			last_value = start_z;
		}
		y += (radius / 2) + 1;
		if (CNC__OutActive == 1 && y < HeightMap->height - radius - 1) {
			for (x = HeightMap->width - radius - 1 - 1; CNC__OutActive == 1 && x >= radius + 1; x--) {
				if (HeightMap->data[(y * HeightMap->width) + x] != 65535 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius, radius) == 0 && (value = CAM__SetMax(CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius), CamSet->z_step)) != 65535) {
				} else {
					value = start_z;
				}
				if (value < last_value) {
					CNC__Go_XY(x, y);
					CNC__Go_Z(value);
				} else if (value > last_value) {
					CNC__Go_XY(x, y);
					CNC__Go_Z(start_z);
				}
				last_value = value;
			}
			if (last_value != start_z) {
				CNC__Go_XY(radius + 1, y);
				CNC__Go_Z(last_value);
				last_value = start_z;
			}
		}
	}
	CNC__Go_Z(start_z);
	CNC__DrillSetSpeed(0);
	return 0;
}

//! CNC-Filler
CAM__HeightMap *CAM__HeightMap_Fill(CAM__HeightMap *HeightMap, CamSetup *CamSet) {
	CAM__HeightMap *HeightMap_New;
	gint x;
	gint y;
	gint value;
	gint radius = (gint)((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / (gfloat)CamSet->zoom) + (gint)(((gfloat)CamSet->axis_z[CamSet->tool].tool.head.radius / 2.0) / (gfloat)CamSet->zoom) + 1;
	if ((HeightMap_New = CAM__HeightMap_create(HeightMap->width, HeightMap->height)) == NULL) {
			return NULL;
	}
	for (y = radius + 1; CNC__OutActive == 1 && y < HeightMap->height - radius - 1; y++) {
		for (x = radius + 1; CNC__OutActive == 1 && x < HeightMap->width - radius - 1; x++) {
			if (HeightMap->data[(y * HeightMap->width) + x] != 65535 && CAM__HeightMap_CheckCircle(HeightMap, x, y, radius, radius) == 0 && (value = CAM__SetMax(CAM__HeightMap_CheckCircleMax(HeightMap, x, y, radius), CamSet->z_step)) != 65535) {
				HeightMap_New->data[(y * HeightMap_New->width) + x] = value;
			} else {
				HeightMap_New->data[(y * HeightMap_New->width) + x] = 65535;
			}
		}
	}
	return HeightMap_New;
}

void CAM__HeightMap_DrawGL(CAM__HeightMap *HeightMap, gchar mode) {
	gint x;
	gint y;
	gint z1;
	gint z2;
	gint z3;
	gint z4;
	gint diff = 10;
	gfloat scale = 10.0;
	gfloat z_scale = app->gl_zzoom;
	scale /= app->hm_zoom;
	if (mode == 0 && HeightMap != NULL) {
		diff = 10;
		glBegin(GL_LINES);
		for (y = 0; y < HeightMap->height; y += diff) {
			for (x = 0; x < HeightMap->width - diff; x += diff) {
				z1 = (gint)((gfloat)HeightMap->data[(y * HeightMap->width) + x]);
				z2 = (gint)((gfloat)HeightMap->data[(y * HeightMap->width) + (x + diff)]);
				if (z1 > app->CamSet->z_start) {
					z1 = app->CamSet->z_start;
				}
				if (z2 > app->CamSet->z_start) {
					z2 = app->CamSet->z_start;
				}
				if (z1 != 0 || z2 != 0) {
					glColor4f((gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, 1.0);
				} else {
					glColor4f(0.0, 0.0, 0.5, 0.3);
				}
				glVertex3f(((gfloat)(x - (HeightMap->width / 2)) / scale), ((gfloat)(y - (HeightMap->width / 2)) / scale), (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
				glVertex3f(((gfloat)(x - (HeightMap->width / 2) + diff) / scale), ((gfloat)(y - (HeightMap->width / 2)) / scale), (gfloat)z2 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
			}
		}
		for (x = 0; x < HeightMap->width; x += diff) {
			for (y = 0; y < HeightMap->height - diff; y += diff) {
				z1 = (gint)((gfloat)HeightMap->data[(y * HeightMap->width) + x]);
				z2 = (gint)((gfloat)HeightMap->data[((y + diff) * HeightMap->width) + x]);
				if (z1 > app->CamSet->z_start) {
					z1 = app->CamSet->z_start;
				}
				if (z2 > app->CamSet->z_start) {
					z2 = app->CamSet->z_start;
				}
				if (z1 != 0 || z2 != 0) {
					glColor4f((gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, 1.0);
				} else {
					glColor4f(0.0, 0.0, 0.5, 0.3);
				}
				glVertex3f(((gfloat)(x - (HeightMap->width / 2)) / scale), ((gfloat)(y - (HeightMap->width / 2)) / scale), (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
				glVertex3f(((gfloat)(x - (HeightMap->width / 2)) / scale), ((gfloat)(y - (HeightMap->width / 2) + diff) / scale), (gfloat)z2 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
			}
		}
		glEnd();
	}
	if (mode == 1 && HeightMap != NULL) {
		if (HeightMap->width < 600 && HeightMap->height < 600) {
			diff = 1;
		} else {
			diff = 5;
		}
		for (x = 0; x < HeightMap->width - diff; x += diff) {
			for (y = 0; y < HeightMap->height - diff; y += diff) {
				z1 = (gint)((gfloat)HeightMap->data[(y * HeightMap->width) + x]);
				z2 = (gint)((gfloat)HeightMap->data[((y) * HeightMap->width) + (x + diff)]);
				z3 = (gint)((gfloat)HeightMap->data[((y + diff) * HeightMap->width) + (x + diff)]);
				z4 = (gint)((gfloat)HeightMap->data[((y + diff) * HeightMap->width) + x]);
				if (z1 > app->CamSet->z_start) {
					z1 = app->CamSet->z_start;
				}
				if (z2 > app->CamSet->z_start) {
					z2 = app->CamSet->z_start;
				}
				if (z3 > app->CamSet->z_start) {
					z3 = app->CamSet->z_start;
				}
				if (z4 > app->CamSet->z_start) {
					z4 = app->CamSet->z_start;
				}
				if (z1 != 0 || z2 != 0 || z3 != 0 || z4 != 0) {
					glBegin(GL_QUADS);
					glColor4f((gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, 1.0);
					glVertex3f(((gfloat)(x - (HeightMap->width / 2)) / scale), ((gfloat)(y - (HeightMap->width / 2)) / scale), (gfloat)z1 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
					glColor4f((gfloat)z2 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z2 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z2 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, 1.0);
					glVertex3f(((gfloat)(x - (HeightMap->width / 2) + diff) / scale), ((gfloat)(y - (HeightMap->width / 2)) / scale), (gfloat)z2 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
					glColor4f((gfloat)z3 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z3 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z3 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, 1.0);
					glVertex3f(((gfloat)(x - (HeightMap->width / 2) + diff) / scale), ((gfloat)(y - (HeightMap->width / 2) + diff) / scale), (gfloat)z3 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
					glColor4f((gfloat)z4 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z4 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, (gfloat)z4 * (10.0 / (gfloat)(app->CamSet->z_start)) / 10.0, 1.0);
					glVertex3f(((gfloat)(x - (HeightMap->width / 2)) / scale), ((gfloat)(y - (HeightMap->width / 2) + diff) / scale), (gfloat)z4 * (10.0 / (gfloat)(app->CamSet->z_start)) * z_scale);
					glEnd();
				}
			}
		}
	}
}
