/** \file cam_imgload.c
 *  \brief CAM-Functions to handle Imagefiles as Heightmap
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <gnome.h>
#include <glade/glade.h>

#include "app.h"
#include "cam.h"
#include "cam_imgload.h"

//! Load Imagefile as new HeightMap
CAM__HeightMap *CAM__Heightmap_LoadImage(gchar *filename, gint min_z, gint max_z, gboolean use_rg, gboolean invert) {
	CAM__HeightMap *HeightMap;
	GdkPixbuf *pixbuf;
	guchar *pixels, *p;
	gint rowstride;
	gint n_channels;
	gint x;
	gint y;
	gint z;
	gint stat;
	gint last_stat;
	gchar text_str[50];
	last_stat = 0;
	max_z -= min_z;
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	n_channels = gdk_pixbuf_get_n_channels (pixbuf);
	rowstride = gdk_pixbuf_get_rowstride (pixbuf);
	pixels = gdk_pixbuf_get_pixels (pixbuf);
	if ((HeightMap = CAM__HeightMap_create(gdk_pixbuf_get_width(pixbuf), gdk_pixbuf_get_height(pixbuf))) == NULL) {
			return NULL;
	}
	HeightMap->use_rg = use_rg;
	for (y = 0; y < HeightMap->height; y++) {
		for (x = 0; x < HeightMap->width; x++) {
			HeightMap->data[(y * HeightMap->width) + x] = 65535;
		}
	}
	for (y = 0; y < HeightMap->height; y++) {
		stat = (gint)((100.0 / (gfloat)HeightMap->height) * (gfloat)y);
		if (stat != last_stat) {
			sprintf(text_str, "Loading Heightmap: %i%%", stat);
			app__set_infobar(text_str);
			app__set_progress(stat);
		}
		while (gtk_events_pending()) {
			  gtk_main_iteration();
		}
		last_stat = stat;
		for (x = 0; x < HeightMap->width; x++) {
			p = pixels + y * rowstride + x * n_channels;
			if (use_rg == TRUE) {
				z = (p[0] * p[1]); // Use Red*Green for the Height-Value(max_z = 256*256 - 1 == 65535)
			} else {
				z = (gint)((gdouble)(65535.0 / 255.0) * (gdouble)p[0]); // Use Red for the Height-Value(max_z = 256 - 1 == 255)
			}
			if (invert == FALSE) {
				z = 65535 - z;
			}
			if (max_z != 65535) {
				if (z > 1) {
					if (z != 65535) {
						z = (gint)((gfloat)z / (65535.0 / (gfloat)max_z));
					}
				}
			}
			if (min_z != 0) {
				z += min_z;
			}
			HeightMap->data[((HeightMap->height - (y + 1)) * HeightMap->width) + (x)] = z;
		}
	}
	app__set_progress(100);
//	gdk_pixbuf_unref(pixbuf);
	return HeightMap;
}

//! Save HeightMap as XPM-Image
void CAM__Heightmap_SaveXPM(CAM__HeightMap *HeightMap, gchar *filename) {
	FILE *file;
	gint x;
	gint y;
	gint stat;
	gint last_stat;
	gchar text_str[50];
	last_stat = 0;
	if ((file = fopen(filename, "w")) != NULL) {
		fprintf(file, "/* XPM */\n");
		fprintf(file, "static char * CAM__XMP_pm[] = {\n");
		fprintf(file, "/* width height ncolors cpp [x_hot y_hot] */\n");
		fprintf(file, "\"%i %i 3 1 0 0\",\n", HeightMap->width, HeightMap->height);
		fprintf(file, "/* colors */\n");
		fprintf(file, "\"/ c black\",\n");
		fprintf(file, "\"_ c gray\",\n");
		fprintf(file, "\"- c white\",\n");
		fprintf(file, "/* pixels */\"\n");
		for (y = 0; y < HeightMap->height; y += 1) {
			stat = (gint)((100.0 / (gfloat)HeightMap->height) * (gfloat)y);
			if (stat != last_stat) {
				sprintf(text_str, "save Heightmap: %i%%", stat);
				app__set_infobar(text_str);
			}
			while (gtk_events_pending()) {
				  gtk_main_iteration();
			}
			last_stat = stat;
			for (x = 0; x < HeightMap->width; x += 1) {
				if (CAM__HeightMap_CheckPixel(HeightMap, x, y) > 0) {
					fprintf(file, "-");
				} else {
					fprintf(file, "/");
				}
			}
			fprintf(file, "\",\n\"");
		}
		fprintf(file, "\"};\n");
		fclose(file);
	}
}

void CAM__Heightmap_SaveTGA(CAM__HeightMap *HeightMap, gchar *filename) {
	FILE *f = NULL;
	gint x, y, z;
	gint stat;
	gint last_stat;
	gchar text_str[50];
	last_stat = 0;
	if ((f = fopen( filename, "w" )) != NULL) {
		fputc (0x00, f);	/* ID Length, 0 => No ID	*/
		fputc (0x00, f);	/* Color Map Type, 0 => No color map included	*/
		fputc (0x02, f);	/* Image Type, 2 => Uncompressed, True-color Image */
		fputc (0x00, f);	/* Next five bytes are about the color map entries */
		fputc (0x00, f);	/* 2 bytes Index, 2 bytes length, 1 byte size */
		fputc (0x00, f);
		fputc (0x00, f);
		fputc (0x00, f);
		fputc (0x00, f);	/* X-origin of Image	*/
		fputc (0x00, f);
		fputc (0x00, f);	/* Y-origin of Image	*/
		fputc (0x00, f);
		fputc (HeightMap->width & 0xff, f);      /* Image Width	*/
		fputc ((HeightMap->width>>8) & 0xff, f);
		fputc (HeightMap->height & 0xff, f);     /* Image Height	*/
		fputc ((HeightMap->height>>8) & 0xff, f);
		fputc (0x18, f);		/* Pixel Depth, 0x18 => 24 Bits	*/
		fputc (0x20, f);		/* Image Descriptor	*/
		fclose(f);
		f = fopen( filename, "ab" );  /* reopen in binary append mode */
		for (y = HeightMap->height - 1; y >= 0; y--) {
			stat = 100 - (gint)((100.0 / (gfloat)HeightMap->height) * (gfloat)y);
			if (stat != last_stat) {
				sprintf(text_str, "save Heightmap: %i%%", stat);
				app__set_infobar(text_str);
			}
			while (gtk_events_pending()) {
				  gtk_main_iteration();
			}
			last_stat = stat;
			//fprintf(stderr, "%i\n", 100 - (100 / height * y));
			for (x = 0; x < HeightMap->width; x++) {
				z = 65535 - HeightMap->data[(y * HeightMap->width) + x];
				z = z / 256;
				fputc(z, f);
				fputc(z, f);
				fputc(z, f);
			}
		}
		fclose(f);
	}
}
